@@ -1,10 +1,12 @@
+from datetime import datetime
 from enum import Enum
 import json
+import logging
 from pydantic import BaseModel, Field, validate_arguments
 import requests
-from urllib.parse import quote_plus
+import time
 from typing import Any, Dict, List, Optional
-from urllib.parse import urlparse
+from urllib.parse import quote_plus, urlparse
 
 
 class FilePath(BaseModel):
@@ -14,13 +16,21 @@ class FilePath(BaseModel):
 
 
 class FieldType(str, Enum):
-    string = "STRING"
+    text = "STRING"
     number = "NUMBER"
     bool = "BOOL"
     timestamp = "TIMESTAMP"
     entity = "ENTITY"
 
 
+class FieldSpec(BaseModel):
+    field: str  # This is the field's name
+    type: FieldType
+    expression: Optional[str]
+    path: Optional[List[str]]
+    isList: Optional[bool]
+
+
 class InferredFieldType(Enum):
     text = {"expression": "`text_string-dev-1`(File.text)", "type": "STRING"}
     number = {"expression": "`text_number-dev-1`(File.text)", "type": "NUMBER"}
@@ -31,13 +41,23 @@ class InferredFieldType(Enum):
         "isList": True,
     }
 
+    @property
+    def expr(self):
+        return self.value["expression"]
 
-class FieldSpec(BaseModel):
-    field: str  # This is the field's name
-    type: FieldType
-    expression: Optional[str]
-    path: Optional[List[str]]
-    isList: Optional[bool]
+    @classmethod
+    def match_trainer(cls, trainer: str):
+        ret = None
+        for x in cls:
+            if trainer in x.expr:
+                assert ret is None, "Matched multiple trainers: %s and %s" % (ret, x)
+                ret = x
+        assert ret is not None, "Unknown trainer: %s" % (trainer)
+        return ret
+
+    @validate_arguments
+    def build_field_spec(self, field_name: str, path: List[str] = []) -> FieldSpec:
+        return FieldSpec(field=field_name, path=path, **self.value)
 
 
 class InvalidRequest(Exception):
@@ -62,6 +82,11 @@ class ResourceType(str, Enum):
     collection = "collection"
 
 
+@validate_arguments
+def parse_date(s: str) -> datetime:
+    return datetime.strptime(s, "%Y-%m-%d")
+
+
 class Impira:
     def __init__(
         self, org_name, api_token, base_url="https://app.impira.com", ping=True
@@ -169,6 +194,19 @@ class Impira:
         if not resp.ok:
             raise APIError(resp)
 
+    @validate_arguments
+    def create_fields(self, collection_id: str, field_specs: List[FieldSpec]):
+        resp = requests.post(
+            urljoin(
+                self.api_url, "schema/ecs/file_collections::%s/fields" % (collection_id)
+            ),
+            headers=self.headers,
+            json=[dict(field_spec) for field_spec in field_specs],
+        )
+
+        if not resp.ok:
+            raise APIError(resp)
+
     @validate_arguments
     def create_inferred_field(
         self,
@@ -177,7 +215,9 @@ class Impira:
         inferred_field_type: InferredFieldType,
         path: List[str] = [],
     ):
-        field_spec = FieldSpec(field=field_name, path=path, **inferred_field_type.value)
+        field_spec = inferred_field_type.build_field_spec(
+            field_name=field_name, path=path
+        )
         return self.create_field(collection_id, field_spec)
 
     @validate_arguments
@@ -219,7 +259,7 @@ class Impira:
 
     @validate_arguments
     def query(
-        self, query: str, mode: str = "iql", cursor: str = None, timeout: str = None
+        self, query: str, mode: str = "iql", cursor: str = None, timeout: int = None
     ):
         args = {"query": query}
         if cursor is not None:
@@ -231,7 +271,7 @@ class Impira:
         resp = requests.post(
             urljoin(self.api_url, mode),
             headers=self.headers,
-            json={"query": query},
+            json=args,
         )
         if not resp.ok:
             raise APIError(resp)
@@ -247,30 +287,36 @@ class Impira:
 
     @validate_arguments
     def _upload_multipart(self, collection_id: str, files: List[FilePath]):
-        for f in files:
-            if f.uid is not None:
-                raise InvalidRequest(
-                    "Unsupported: specifying a UID in a multi-part file upload (%s)"
-                    % (f.uid)
-                )
+        for i in range(60):
+            for f in files:
+                if f.uid is not None:
+                    raise InvalidRequest(
+                        "Unsupported: specifying a UID in a multi-part file upload (%s)"
+                        % (f.uid)
+                    )
 
-        files_body = [
-            t
-            for f in files
-            for t in [
-                ("file", open(f.path, "rb")),
-                ("data", json.dumps(_build_file_object(f.name, None, f.uid))),
+            files_body = [
+                t
+                for f in files
+                for t in [
+                    ("file", open(f.path, "rb")),
+                    ("data", json.dumps(_build_file_object(f.name, None, f.uid))),
+                ]
             ]
-        ]
-        resp = requests.post(
-            self._build_collection_url(collection_id, use_async=True),
-            headers=self.headers,
-            files=tuple(files_body),
-        )
-        if not resp.ok:
-            raise APIError(resp)
-
-        return resp.json()["uids"]
+            resp = requests.post(
+                self._build_collection_url(collection_id, use_async=True),
+                headers=self.headers,
+                files=tuple(files_body),
+            )
+            if resp.status_code == 429:
+                logging.warning(
+                    "Sleeping for 2 seconds and then retrying multi-part upload..."
+                )
+                time.sleep(2)
+            elif not resp.ok:
+                raise APIError(resp)
+            else:
+                return resp.json()["uids"]
 
     @validate_arguments
     def _upload_url(self, collection_id: str, files: List[FilePath]):
